# Contributing to Reselect

:rocket: Awesome, thanks for wanting to contribute to Reselect! :rocket:

The following is a set of guidelines for contributing to Reselect and its packages, which are hosted at [Reselect](https://github.com/reselect/Reselect) on GitHub.

[How Can I Contribute?](#how-can-i-contribute-book)
- [Reporting Bugs](#reporting-bugs-beetle)
- [Your First Code Contribution](#your-first-code-contribution-point_up)
- [Pull Requests](#pull-requests)

[Styleguides](#styleguides)
- [Git Commit Messages](#git-commit-messages)
- [Specs Styleguide](#specs-styleguide)
- [Documentation Styleguide](#documentation-styleguide)

## How Can I Contribute :book:

### Reporting Bugs :beetle:

Before creating a bug report, please check [this list](https://github.com/reselect/Reselect/issues?q=is%3Aopen+is%3Aissue+label%3Abug). When submitting a bug report, please include as many details as possible.

### Your First Code Contribution :point_up:

Looking for a good place to start contributing? Look through these `beginner` and `help wanted` issues:
- [Beginner Issues]() - issues that should only require a few lines of code and minimal testing.
- [Help Wanted Issues](https://github.com/reselect/Reselect/labels/help%20wanted) - issues that are more involved than `beginner` issues.

### Pull Requests

- Include screenshots and/or animated GIFs whenever possible.
- Follow all style guides
- Include well-worded and structured tests
- End files with new lines

## Styleguides

### Git Commit Messages

Structure your git commit message as per the Git documentation.

From: [[http://git-scm.com/book/ch5-2.html]]

> Short (50 chars or less) summary of changes

> More detailed explanatory text, if necessary.  Wrap it to about 72 characters or so.  In some contexts, the first line is treated as the subject of an email and the rest of the text as the body.  The blank line separating the summary from the body is critical (unless you omit the body entirely); tools like rebase can get confused if you run the two together.

> Further paragraphs come after blank lines.

>  - Bullet points are okay, too

>  - Typically a hyphen or asterisk is used for the bullet, preceded by a single space, with blank lines in between, but conventions vary here

TL:DR;

1. Separate the subject line from the body
2. Include the type of change in your subject i.e "Test", "Feat"
3. Include the branch or feature in brackets
4. Limit the subject line to 50 characters
5. Capitalize the subject line
6. Don't end the subject line with a period
7. Wrap the message body at 72 characters
8. Use an imperative mood i.e "Fix", "Change", "Add" vs "Fixed", "Changed"
9. Explain what and why the change was made *not* how it was changed
10. Split up your commits if it includes several changes to logic

#### Example

> Test(reselect-choices): Adding reselect tests

> Add basic test to ensure reselect-choices outputs correct values.

### Documentation Styleguide

TODO: Documentation Styleguide
