Reselect.service('LazyScroller', ['LazyContainer', '$compile', function (LazyContainer, $compile) {
    var defaultOptions = {
        scopeName: '$choice'
    }

    var LazyScroller = function ($scope, options) {
        var self = this

        self.options = angular.extend({}, defaultOptions, options)

        self.$scope = $scope

        self.$container = self.options.container
        self.$list = self.options.list

        self.choices = []

        self.lazyContainers = []
        self.numLazyContainers = Math.ceil((self.options.listHeight) / self.options.choiceHeight) + 2

        self.lastCheck = null
        self.lastScrollTop = null
    }

    LazyScroller.prototype.renderContainer = function () {
        var self = this

        // Set the max height of the dropdown container
        // var optionsHeight = $Reselect.choices.length * self.choiceHeight;
        var optionsHeight = self.choices.length * self.options.choiceHeight
        var containerHeight = (optionsHeight > self.options.listHeight) ? self.options.listHeight : optionsHeight

        self.$container.css('height', (containerHeight || self.options.choiceHeight) + 'px')

        // Simulate the scrollbar with the estimated height for the number of choices
        self.$list.css('height', optionsHeight + 'px')

        return {
            choiceHeight: optionsHeight,
            containerHeight: containerHeight
        }
    }

    LazyScroller.prototype.bindScroll = function () {
        var self = this

        self.$container.on('scroll', function () {
            window.requestAnimationFrame(function () {
                self._calculateLazyRender()

                self.$scope.$apply()
            })
        })
    }

    LazyScroller.prototype._shouldRender = function (scrollTop) {
        var self = this

        return !(typeof self.lastCheck === 'number' &&
            (
                scrollTop <= self.lastCheck + (self.options.choiceHeight - (self.lastCheck % self.options.choiceHeight)) && //
                scrollTop >= self.lastCheck - (self.lastCheck % self.options.choiceHeight) //
            ))
    }

    LazyScroller.prototype._calculateLazyRender = function (force) {
        var self = this

        var scrollTop = self.$container[0].scrollTop

        self.lastScrollTop = scrollTop

        // A Check to throttle amounts of calculation by setting a threshold
        // The list is due to recalculation only if the differences of scrollTop and lastCheck is greater than a choiceHeight
        if (force !== true) {
            if (!self._shouldRender(scrollTop)) {
                return
            }
        }

        var activeContainers = []
        var inactiveContainers = []

        angular.forEach(self.lazyContainers, function (lazyContainer, index) {
            var choiceTop = (lazyContainer.index) * self.options.choiceHeight || 0

            if (force === true) {
                lazyContainer.index = null
            }

            // Check if the container is visible
            if (lazyContainer.index === null || choiceTop < scrollTop - self.options.choiceHeight || choiceTop > scrollTop + self.options.listHeight + self.options.choiceHeight) {
                lazyContainer.element.addClass('inactive').removeClass('active')
                inactiveContainers.push(lazyContainer)
            } else {
                lazyContainer.element.addClass('active').removeClass('inactive')
                activeContainers.push(lazyContainer)
            }
        })

        var indexInDisplay = activeContainers.map(function (container) {
            return container.index
        })

        // Get the start and end index of all the choices that should be in the viewport at the current scroll position
        var indexToRenderStart = Math.floor(scrollTop / self.options.choiceHeight)
        indexToRenderStart = indexToRenderStart < 0 ? 0 : indexToRenderStart

        var indexToRenderEnd = Math.ceil((scrollTop + self.options.listHeight) / self.options.choiceHeight)
        indexToRenderEnd = indexToRenderEnd >= self.choices.length ? self.choices.length : indexToRenderEnd

        // Start rendering all missing indexs that is not in the viewport
        for (var i = indexToRenderStart; i < indexToRenderEnd; i++) {
            if (indexInDisplay.indexOf(i) >= 0) {
                continue
            } else {
                // Get the next available lazy container
                var container = inactiveContainers.shift()

                if (container) {
                    container.element.addClass('active').removeClass('inactive')

                    container.index = i
                    container.render(self.options.choiceHeight)

                    angular.extend(container.scope, {
                        $containerId: container.containerId,
                        $index: i,
                        cssClass: self.choices[i] ? self.choices[i].class : ''
                    })

                    container.scope.$group = false

                    if (self.choices[i] && self.choices[i].$sticky === true) {
                        container.scope[self.options.scopeName] = self.choices[i].value
                        container.scope.$onClick = self.choices[i].onClick
                        container.scope.$sticky = self.choices[i].$sticky
                        container.scope.$stickyContent = self.choices[i].$stickyContent
                    } else if (self.choices[i] && self.choices[i].$$group) {
                        container.scope.$group = true
                        container.scope.$groupContent = self.choices[i].$$group
                    } else {
                        container.scope[self.options.scopeName] = self.choices[i]
                        container.scope.$sticky = false
                    }
                }
            }
        }

        // self.lastCheck = Math.floor(scrollTop/self.options.choiceHeight) * self.options.choiceHeight || null;
        self.lastCheck = scrollTop || null
    }

    LazyScroller.prototype.initialize = function (tpl) {
        var self = this

        for (var i = 0; i < self.numLazyContainers; i++) {
            var $choice = tpl.clone()

            // HACK
            var lazyScope = self.$scope.$new()
            lazyScope.$options = self.$scope.$options
            lazyScope[self.options.scopeName] = {}

            $compile($choice)(lazyScope)

            self.lazyContainers.push(new LazyContainer({
                containerId: i,
                element: $choice,
                scope: lazyScope
            }))

            self.$list.append($choice)
        }

        self.bindScroll()
    }

    return LazyScroller
}])

Reselect.service('LazyContainer', [function () {
    var LazyContainer = function (options) {
        var self = this

        self.containerId = null
        self.element = null
        self.index = null
        self.scope = null

        angular.extend(self, options)
    }

    LazyContainer.prototype.render = function (containerHeight) {
        var self = this

        if (!self.element && self.index === null) {
            return
        }

        self.element.css('top', (self.index * containerHeight) + 'px')
    }

    return LazyContainer
}])
